/*
 * @(#)cdemostate.c    1.00   02/21/02
 *
 * Copyright (c) 2002-Present Agilis Software LLC, Santa Clara, CA, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Agilis Software LLC ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only
 * in accordance with the terms of the license agreement you entered
 * into with Agilis Software LLC.
 *
 * DISCLAIMER: this sample code is provided as-is.
 */

/*
 * This demo illustrates how to securely manage state in an
 * EasyLicenser-enabled application using the secure state management 
 * API's for maintaining quota consumption and other license-related 
 * application state.
 *
 * Running the demo once is equivalent to your end customer running your 
 * license-protected application at their site and consuming a specified 
 * amount of quota at their site.  The quota consumption is specified
 * as a command line argument.  Initialization of state accomplished
 * by running the demo with a quota consumption value of 0.
 * For illustration purposes, the demo saves state in a file in
 * the current directory.
 *
 * Features illustrated with this demo:
 * 1. How to use the "checkLicenseKeySecure_s" API call and the key cookie
 *    mechanism to manage and update license state including the built-in
 *    quota management and the extensibility mechanism to
 *    manage application-specific state.
 * 2. How to utilize the application-password security mechanism.
 * 3. How to protect against system clock changes without writing code.
 *
 * For examples on using custom key handlers for extensibility,
 * as well as using digital signatures and other anti-spoofing measures,
 * see the accompanying "cdemosecure.cpp" demo application.
 *
 * Some important points to note about this demo:
 * 1. The code that is directly relevant to using the EasyLicenser API's
 *    is indicated as such in comments.  The rest of the code exists in
 *    order to have a functioning self-contained program.
 * 2. All code pertaining to reading the key from an EasyLicenser file is
 *    for functional purposes only.  It is not meant to suggest a specific
 *    technique for managing the key in your application.  Nor is it
 *    a part of the EasyLicenser product.  In particular, note that your
 *    actual application will typically have access to just the license key
 *    without any accompanying metadata about the key.
 * 3. The demo will work with keys generated with EasyLicenser 2.0 as well
 *    as prior versions.  The EasyLicenser 2.0 run time library is required.
 * 4. This demo performs automatic state management in a file that persists
 *    past its execution.  Therefore, each successive run will have a 
 *    different outcome and eventually you should encounter an "exceeded
 *    quota" error if you supplied a quota limited key.
 * 5. Unlike the basic "C_Demo.c", this demo processes only the first key
 *    in its input file.  Therefore, provide a file containing only one key.
 *    The key need not be quota limited.  If it isn't, the key cookie will
 *    still manage a use count, which can be useful in an actual application.
 */

#define __STRICT_ANSI__ 1

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <ctype.h>

#include <ezLicenseInfo.h>

/* File types */
#define EZLM_FILE_TYPE_UNKNOWN	0
#define EZLM_FILE_TYPE_TXT		1
#define EZLM_FILE_TYPE_XML		2
#define EZLM_FILE_TYPE_CSV		3
#define EZLM_FILE_TYPE_TAB		4
#define EZLM_FILE_TYPE_INI		5
#define EZLM_FILE_TYPE_KEY		6

void CALLING_CONVENTION displayKeyCookie(
							const char *keyCookie, 
							const char *productName, 
							const char *appPassword);
void CALLING_CONVENTION updateKeyCookieInFile(const char *keyCookie);
void CALLING_CONVENTION getKeyCookieFromFile(char keyCookie[], size_t keyCookieBufsize);
void CALLING_CONVENTION displayWarnings(int iWarningsBitMap);
void CALLING_CONVENTION displayError(int errorCode);
void CALLING_CONVENTION displayLicenseParams(void* ezLicInstance);
int CALLING_CONVENTION getFileType(const char *fname);
int CALLING_CONVENTION getKey(char *filelines[], int numlines,
							  int currLine, const int fileType,
							  char keybuf[], size_t keybufsize);
int CALLING_CONVENTION readFile(const char *filename, char *filelines[], int maxfilelines);
int CALLING_CONVENTION strip(const char *instr, char retbuf[], size_t retbufsize);
void makeDateStr(time_t ltime, char dateStrBuf[], size_t dateStrBufsize);

/**
 * main entry point:
 * Accepts three mandatory command line arguments:
 * 1. Name of a text file containing a quota-limited license key.
 * 2. Expected user / host name that should match what's in the key.
 *    If the key is known to have been generated with the "enforce OS
 *    user / host" flag, you can specify the special string "null" to
 *    signify a null value.
 * 3. Quota increment: amount by which to increment quota consumption.
 *    To reset the saved state, specify a value of 0 to simulate
 *    product install time.
 * and two optional arguments (either none or both must be specified):
 * 4. Case-sensitive product name that was specified at the time of key creation,
 *    if the product had an associated application password
 * 5. Product's associated password's case-sensitive public key, available with
 *    the product definition and also obtainable with the "makepubkey" Java utility
 *    located at ezlic2.0/util.
 */
int main(int argc, char **argv)
{
	FILE* fptr = 0;
	char 	*strLicenseKeyFileName = NULL;
	char 	*strLicenseKey = NULL;
	int		iDaysLeftThreshold = 7;
	long	lQuotaUsageToDate = 0;
	long	lQuotaThreshold = 10;
	int		iDaysGracePeriod = 365; 
	long	lQuotaGrace = 100;
	long	lCurrUsageValue = 0;
	int		iStatusCode = 0;
	char	*strUserHostName;
	void *ezLicInfoInstance = 0;
	char *filelines[200];
	int numlines = 0;
	char keybuf[MAX_ENCRYPTED_LICENSE_TEXT_LENGTH];
	char keyCookie[MAX_KEY_COOKIE_SIZE];
	char newKeyCookie[MAX_KEY_COOKIE_SIZE];
	char appState[MAX_APP_STATE_SIZE];
	char *appPwd = NULL;
	char *productName = NULL;
	int fileType = EZLM_FILE_TYPE_UNKNOWN;
	int currLine = 0;
	int ret = 0;

    char    ezlmMajorVersion[MAX_VERSION_INFO_LENGTH];
    char    ezlmMinorVersion[MAX_VERSION_INFO_LENGTH];

	if (argc < 4 || argc == 5) {
usage:
		printf("\n*** usage: %s key-text-file-name user-or-host-name quota-incr [product-name application-password-public-key] ***", argv[0]);
		printf("\n    Example: %s keypub.csv user1 1\n", argv[0]);
		exit(2);
	}

	strLicenseKeyFileName = argv[1];
	strUserHostName = argv[2];
	lQuotaUsageToDate = atol(argv[3]);
	if (lQuotaUsageToDate < 0) {
		printf("*** Invalid quota increment value '%s' specified - must be integer >= 0\n", argv[3]);
		goto usage;
	}
	if (argc > 4) {
		productName = argv[4];
		appPwd = argv[5];
	}

	printf("\nLicense State Demo.  Running in %s mode.\n", (lQuotaUsageToDate == 0? "application initialization":"application execution"));
    printf("\nRuntime library version information:\n");
    if (getEzlmVersionInfoStr(ezlmMajorVersion, sizeof(ezlmMajorVersion),
                              ezlmMinorVersion, sizeof(ezlmMinorVersion)) < 0) {
        printf("*** Error getting EasyLicenser runtime version information ***\n");
        exit (2);
    }
    printf("  Major Version: %s, Minor Version: %s\n", ezlmMajorVersion, ezlmMinorVersion);

	printf("\nCommand line arguments:\n");
	printf("  License key file:          %s\n", strLicenseKeyFileName);
	printf("  User / Host name:          %s\n", strUserHostName);
	printf("  Quota increment:           %ld\n", lQuotaUsageToDate);
	printf("  Product name:              %s\n", (productName == NULL ? "<not specified>":productName));
	printf("  App Passwd Public Key:     %s\n", (appPwd == NULL ? "<not specified>":appPwd));

	if (!lQuotaUsageToDate) {
		/* Simulate product and key installation: initialize key cookie and exit. */

		/* 
		 * Set app state to current date string, for illustration purposes.  In a real
		 * application, you can use the application state to store arbitrary licensing
		 * related information.
		 */
		makeDateStr(0, appState, sizeof(appState));
		/* EasyLicenser API call: create an initial key cookie at product install time */
		if ((ret = createKeyCookieSeed_c(productName, appPwd, appState, FALSE,
								  	     keyCookie, sizeof(keyCookie))) < 0) {
			printf("*** Error creating key cookie seed ***\n");
			displayError(ret);
			exit(2);
		}
		updateKeyCookieInFile(keyCookie);
		printf("Initialized license state.  Exiting.\n");
		exit(0);
	}

	/* Application run time simulation: get the last key cookie from persistent store */
	getKeyCookieFromFile(keyCookie, sizeof(keyCookie));
	printf("\nDump of key cookie before license key check:\n");
	displayKeyCookie(keyCookie, productName, appPwd);

	if (strcmp(strUserHostName,"null")==0)
		strUserHostName = NULL;

	fileType = getFileType(strLicenseKeyFileName);

	numlines = 
		readFile(strLicenseKeyFileName, filelines, 1000);
	if (numlines < 0) {
		printf("*** Cannot process license key file due to errors.  Aborting. ***\n");
		exit(1);
	} else if (!numlines) {
		printf("License key file is empty.  Nothing to do.  Exiting.\n");
		exit(0);
	}

	printf("\nChecking first license key in file...\n");
	strLicenseKey = &keybuf[0];
	if (getKey(filelines, numlines, 0, fileType, keybuf, sizeof(keybuf)) <= 0)
		exit(1);

	printf("\nLicense key:\n");
	printf("[%s]\n\n", keybuf);

	/* 
	 * Update app state to current date string, for illustration purposes.  In a real
	 * application, you can use the application state to store arbitrary licensing
	 * related information.
	 */
	makeDateStr(0, appState, sizeof(appState));

	/* EasyLicenser API call (required): allocate a license context. */
	ezLicInfoInstance = allocLicenseInfo();
	/* EasyLicenser API call (required): perform the license key check. */
	iStatusCode = checkLicenseKeySecure_s(
					ezLicInfoInstance,	/* license context */
					keybuf, 			/* the license key string */
					keyCookie,			/* the last-obtained key cookie */
					newKeyCookie,		/* buffer to receive updated  */
										/* key cookie */
					sizeof(newKeyCookie),  /* output buffer size */
					0, 0,				/* no custom key handler  */
										/* for this simple example */
					iDaysLeftThreshold, /* threshold days to trigger  */
										/* "about to expire" warning */
					-lQuotaUsageToDate,	/* quota increment - note it has to */
										/* be specified as a -ve incremental # */
										/* in order for the automatic state */
										/* management to kick in.  Note also */
										/* that a -ve value implies an increment */
					lQuotaThreshold, 	/* threshold balance quota to */
										/* trigger "about to exhaust */
										/* quota" warning */
					iDaysGracePeriod,	/* allow license key check to */
										/* succeed with warning instead */
										/* of error within these many */
										/* days past expiration date. */
					lQuotaGrace,		/* allow license key check to */
										/* succeed with warning instead */
										/* of error within these many */
										/* uses past quota limit. */
					strUserHostName, 	/* user / host name - can be null */
										/* if OS enforcement is in effect. */
					lCurrUsageValue,	/* current use level (note: different */
										/* from quota which is cumulative */
										/* and can use different units) - */
										/* checked against limit in key. */
										/* applies only to non-single-user */
										/* licenses. */
					productName,		/* name of product for which key */
										/* was generated */
					appPwd,				/* application password public key, */
										/* if the product had an application */
										/* password specified when the key was */
										/* generated. */
					appState);			/* new application state to save in */
										/* key cookie. */

	if (iStatusCode < 0) {
		/* license check failed: report the error condition */
		displayError(iStatusCode);
	} else {	
		/* license check succeeded.  If the return value is > 0, */
		/* it means there were warnings issued - report these */
		printf("    License key validation successful%s.\n\n", (iStatusCode?" with warnings":""));
		/* introspect the entire license key contents - */
		/* not something you'd do in an actual application */
		displayLicenseParams(ezLicInfoInstance);
		displayWarnings(iStatusCode);
		printf("\nDump of new key cookie after license key check:\n");
		displayKeyCookie(newKeyCookie, productName, appPwd);
		updateKeyCookieInFile(newKeyCookie);
	}
	/* EasyLicenser API call (required): release the license context. */
	freeLicenseInfo(ezLicInfoInstance);
	printf("\nEnd of demo.\n");
	return 0;
}

/*
 * Display state contents in the encrypted key cookie.  Note that 
 * you can introspect a key cookie but it can't be created / 
 * updated outside the EasyLicenser API's.
 * In an actual application, you may be interested in introspecting
 * an application-specific subset of the cookie information, none
 * in the simple case.
 */
void CALLING_CONVENTION displayKeyCookie(
							const char *keyCookie, 
							const char *productName, 
							const char *appPassword)
{
	char strBuffer[MAX_APP_STATE_SIZE];
	long quotaToDate = -1;
	time_t updateTime = 0;
	printf("  Key cookie:\n[%s]\n", keyCookie);
	/* EasyLicenser API call: get application state from key cookie */
	if (getKeyCookieAppState(keyCookie, productName, appPassword,
							 strBuffer, sizeof(strBuffer)) < 0) {
		printf("*** Error getting application state ***\n");
	} else
		printf("    Application state: [%s]\n", strBuffer);

	/* EasyLicenser API call: get quota-consumed-to-date state from key cookie */
	if (getKeyCookieQuotaToDate(keyCookie, productName, appPassword,
							    &quotaToDate) < 0) {
		printf("*** Error getting quota-to-date ***\n");
	} else
		printf("    Quota-to-date: %ld\n", quotaToDate);

	/* EasyLicenser API call: get cookie update timestamp from key cookie */
	if (getKeyCookieTs(keyCookie, productName, appPassword,
					   &updateTime) < 0) {
		printf("*** Error getting update timestamp ***\n");
	} else {
		makeDateStr(updateTime, strBuffer, sizeof(strBuffer));
		printf("    Update date/time (GMT): %s (timestamp: %ld)\n", strBuffer, updateTime);
	}

}


/*
 * Update key cookie in file.  For demo purposes only.  In an actual
 * application, you'd persist your key cookie in an application-
 * specific manner.
 */
void CALLING_CONVENTION updateKeyCookieInFile(const char *keyCookie)
{
	int ret = 0;
	FILE *fh = fopen("mycookiefile", "w");
	if (!fh) {
		printf("*** Error opening key cookie file for write ***\n");
		exit(2);
	}
	ret = fputs(keyCookie, fh);
	if (ret < 0) {
		printf("*** Error writing key cookie to file ***\n");
	}
	fclose(fh);
	if (ret < 0)
		exit(2);
}

/*
 * Get key cookie from file.  For demo purposes only.  In an actual
 * application, you'd persist your key cookie in an application-
 * specific manner.
 */
void CALLING_CONVENTION getKeyCookieFromFile(char keyCookie[], size_t keyCookieBufsize)
{
	FILE *fh = fopen("mycookiefile", "r");
	if (!fh) {
		printf("*** File for key cookie does not exist ***\n");
		exit(2);
	}
	fgets(keyCookie, keyCookieBufsize, fh);
	fclose(fh);
	if (!keyCookie[0]) {
		printf("*** File for key cookie is empty ***\n");
		exit(2);
	}
}


/*
*	Display warnings if any.  Warning codes are defined in <ezLicStatusCodes.h>
*	@param iWarningsBitMap : Warning bitMap returned from license key API check.
*/
void CALLING_CONVENTION displayWarnings(int iWarningsBitMap)
{
	if (!iWarningsBitMap)
		return;
	printf("        Warning codes:\n");
    if (iWarningsBitMap & EZLIC_WARN_EXPIRING)
        printf("          Time limited license key is about to expire.\n");
	if (iWarningsBitMap & EZLIC_WARN_QUOTA)
        printf("          Quota limited license key is about to expire.\n");
	if (iWarningsBitMap & EZLIC_WARN_EXPIRED)
        printf("          Time limited license has expired and is now in grace period.\n");
	if (iWarningsBitMap & EZLIC_WARN_QUOTA_EXCEEDED)
        printf("          Quota limited license key has expired and is now in grace period.\n");
	if (iWarningsBitMap & EZLIC_WARN_EXP_GRACE_TRUNC)
		printf("          Expiration Days Grace In API Call Too Large - Truncated To 5 / 25 pct Of Lifespan\n");
	if (iWarningsBitMap & EZLIC_WARN_QUOTA_GRACE_TRUNC)
		printf("          Quota Grace In API Call Too Large - Truncated To 5 / 25 pct Of Original Quota\n");
}


/*
*	Display the meaning of the -ve error code returned from the
*   license key API check.  Error codes are defined in <ezLicStatusCodes.h>
*/
void CALLING_CONVENTION displayError(int errorCode)
{
	if (!errorCode)
		return;

	printf("*** ");
	switch(errorCode) {
	case     EZLIC_EXC_BASE:
			printf("Error: General license validation failure. Please check your license key.\n");
			break;
	case     EZLIC_EXC_MALFORMED_KEY:
			printf("Error: the license key is invalid.\n");
			break;
	case     EZLIC_EXC_EXPIRED:
			printf("Error: The license has expired past the grace period.\n");
			break;
	case     EZLIC_EXC_EXCEEDED_LIMITS:
			printf("Error: The licensed user / server / CPU License / quota limit in the key has been exceeded.\n");
			break;
	case     EZLIC_EXC_INVALID_HOST_USER_NET:
			printf("Error: The user / machine name in the license key does not match your environment / parameter.\n");
			break;
	case     EZLIC_EXC_INVALID_CUSTOM_KEY:
			printf("Error: The custom key contained in the license key is invalid.\n");
			break;
	case     EZLIC_EXC_UNIMPLEMENTED:
			printf("Error: The feature you are using is not implemented.\n");
			break;
	case     EZLIC_EXC_GEN:
			printf("Error: Key generation execption (you shouldn't encounter this condition).\n");
			break;
	case     EZLIC_EXC_VENDOR:
			printf("Error: Vendor license key exception (you shouldn't encounter this condition).\n");
			break;
	case     EZLIC_EXC_INTERNAL:
			printf("Error: internal error in EasyLicenser - please submit a bug report.\n");
			break;
	case     EZLIC_EXC_OS:
			printf("Error: An operating system error occurred - please check your machine configuration.\n");
			break;
	case     EZLIC_EXC_CONFIG:
			printf("Error: exception in system configuration - please check your EasyLicenser installation and system clock.\n");
			break;
	case     EZLIC_EXC_UNLICENSED_OPTION:
			printf("Error: you attempted to perform an unlicensed operation.\n");
			break;
	case     EZLIC_EXC_USAGE:
			printf("Error: API usage - please check your API call parameters and call sequence.\n");
			break;
	case     EZLIC_EXC_AUTHENT:
			printf("Error: Key / cookie authentication failure - please check your product name and application password.\n");
			break;
	default:
			printf("Error: unknown error code %d.\n", errorCode);
			break;
	}
}

void CALLING_CONVENTION showLicenseModel(int licenseModel,
										 int licenseType)
{
	printf("        License Model: %s\n", 
			(licenseModel == EZLIC_MODEL_USER ? "User":
			 licenseModel == EZLIC_MODEL_SERVER ? "Server":
			 "*** Unknown ***"));
	printf("        License Type: ");
	switch(licenseModel) {
	case EZLIC_MODEL_USER:
		switch(licenseType) {
		case EZLIC_TYPE_U_USER:
			printf("Named User\n");
			break;
		case EZLIC_TYPE_U_NODE:
			printf("Node Locked\n");
			break;
		case EZLIC_TYPE_U_FLOAT:
			printf("Floating User\n");
			break;
		default:
			printf("*** Unknown license type %d ***\n", licenseType);
		}
		break;
	case EZLIC_MODEL_SERVER:
		switch(licenseType) {
		case EZLIC_TYPE_SVR_CONC:
			printf("Concurrent User\n");
			break;
		case EZLIC_TYPE_SVR_CPU:
			printf("CPU Count\n");
			break;
		case EZLIC_TYPE_SVR_MHZ:
			printf("CPU MHZ\n");
			break;
		case EZLIC_TYPE_SVR_NMU:
			printf("Named User\n");
			break;
		default:
			printf("*** Unknown license type %d ***\n", licenseType);
		}
		break;
	default:
		printf("%d\n", licenseModel);
	}
}


/*
* Introspect the license key and display its parameters.
* In an actual application you may introspect specific
* values and act on them instead of displaying them.
* The most common parameter that is post-processed is
* the product options.
*/
void CALLING_CONVENTION displayLicenseParams(
							void* ezLicInstance)
{
	int value = 0;
	long lvalue = 0L;
	long ltime = 0;
	char* strValue = 0;


	printf("    License Key Information:\n");
	printf("        Key version:    ");
	printf((getKeyVersion(ezLicInstance) <= 0 ? "Legacy (1.1 or earlier)\n":"%d\n"),
		   getKeyVersion(ezLicInstance));
	strValue = getChainedKeyId(ezLicInstance);
	if (strValue && strValue[0])
		printf("        Replacement for key with Id: '%s'\n", strValue);

	value = getLicenseModelCode(ezLicInstance);
	showLicenseModel(value, getLicenseTypeCode(ezLicInstance));
	if (getEnforce(ezLicInstance))
		printf("        Operating System check enforced.\n");

	value = getLicenseModeBitmap(ezLicInstance);
	printf("        License Mode Bitmap:    0x%x\n", value);
	if (value & EZLIC_MODE_TIME)
		printf("          TIME LIMITED\n");
	if (value & EZLIC_MODE_OPTIONS)
		printf("          OPTIONS\n");
	if (value & EZLIC_MODE_METERED)
		printf("          METERED\n");
	if (value & EZLIC_MODE_CUSTOM_KEY)
		printf("          CUSTOM KEY\n");
	if (value & EZLIC_MODE_CUSTOM_COOKIE)
		printf("          CUSTOM COOKIE\n");

	lvalue = getUsageValue(ezLicInstance);
	if (lvalue > 0)
		printf("        Usage Limit Value:    %ld\n", lvalue);

	lvalue = getQuotaValue(ezLicInstance);
	if (lvalue > 0)
		printf("        Quota Limited To:    %ld\n", lvalue);

	strValue = getOptions(ezLicInstance);
	if (strValue  && strValue[0])
		printf("        Options:    [%s]\n", strValue);

	if ((ltime = getExpirationTime(ezLicInstance)) > 0) {
		char tmpbuf[100];
		struct tm when;
		time_t t = (time_t) ltime;
		when = *gmtime( &t );
		strcpy(tmpbuf,asctime(&when));
		if (tmpbuf[strlen(tmpbuf)-1]=='\n')
			tmpbuf[strlen(tmpbuf)-1] = '\0';
		printf( "        Time Limited To Expire (GMT):    %s\n", 
				tmpbuf);
	}
   	
	strValue = getUserHostNetName(ezLicInstance);
	if (strValue && strValue[0])
		printf("        User/Host Name:    [%s]\n", strValue);
	strValue = getCustomCookie(ezLicInstance);
	if (strValue && strValue[0])
		printf("        Custom Cookie:    [%s]\n", strValue);
	strValue = getCustomKey(ezLicInstance);
	if (strValue && strValue[0])
		printf("        Custom Key:    [%s]\n", strValue);
	strValue = getAppPwd(ezLicInstance);
	if (strValue && strValue[0])
		printf("        Application Password:    [%s]\n", strValue);
}


/* 
 * The file IO is for convenience purposes only - you'd manage the persistence of 
 * your key according to the requirements of your application.
 */

int CALLING_CONVENTION readFile(const char *filename, char *filelines[], int maxfilelines)
{
	int numlines = 0;
	int totlen = 0;
	int len = 0;
	char linebuf[10240];
	char *linep = NULL;
	char *lp = NULL;
	int trunc = 0;
	FILE *fh = fopen(filename,"r");
	if (!fh) {
		printf("\n*** Cannot open license key file '%s' ***\n",
				filename);
		return -1;
	}
	while (numlines <= maxfilelines &&
	       fgets(linebuf, sizeof(linebuf), fh)) {
		linep = (char *)malloc(sizeof(linebuf));
		lp = &linebuf[0];
		while (*lp && isspace((int)*lp))
			lp++;
		if (*lp) {
			strcpy(linep, lp);
			filelines[numlines++] = linep;
		}
	}
	if ((!trunc) && numlines > maxfilelines)
		printf("\n*** Warning: IO buffer overflow, truncating.\n");
	fclose(fh);
	return numlines;
}


int CALLING_CONVENTION getKey(char *filelines[], int numlines,
							  int currLine, const int fileType,
							  char keybuf[], size_t keybufsize)
{
	const char *keyp = NULL;

	if (currLine >= numlines)
		return 0;

	if (currLine == 0) {
		/* check / adjust for first line */
		switch (fileType) {
		case EZLM_FILE_TYPE_TXT:
			if (strncmp(filelines[currLine++], "License Key List:", strlen("License Key List:")) ||
			    strncmp(filelines[currLine++], "----------------", strlen("----------------"))) {
				printf("*** Not a valid EasyLicenser text export file ***\n");
				exit(2);
			}
			currLine += 3; /* position at "Key Id:" */
			break;
		case EZLM_FILE_TYPE_XML:
			if (strncmp(filelines[currLine++], "<?xml", strlen("<?xml")) ||
			    strncmp(filelines[currLine++], "<key-list>", strlen("<key-list>"))) {
				printf("*** Not a valid EasyLicenser XML export file ***\n");
				exit(2);
			}
		default:
			break;
		}
	} else {
		/* check for EOF */
		switch (fileType) {
		case EZLM_FILE_TYPE_XML:
			if (filelines[currLine][0] && strncmp(filelines[currLine],"</key-list>",strlen("</key-list>"))==0)
				return 0;
		default:
			if (!filelines[currLine][0])
				return 0;
			break;
		}
	}
	/* get key */
	switch (fileType) {
	case EZLM_FILE_TYPE_TXT:
		for (; strncmp(filelines[currLine], "License key:", strlen("License key:")) && 
		       currLine < numlines; 
		     currLine++);
		if (++currLine >= numlines) {
			printf("*** Premature termination of text file ***\n");
			return 0;
		}
		keyp = filelines[currLine];
		strip(keyp, keybuf, keybufsize);
		currLine += 2;
		break;
	case EZLM_FILE_TYPE_XML:
		if (strncmp(filelines[currLine], "<key id=", strlen("<key id="))) {
			printf("*** Invalid XML record - treating as EOF ***\n");
			return 0;
		}
		if (++currLine >= numlines) {
			printf("*** Premature termination of XML file ***\n");
			return 0;
		}
		keyp = filelines[currLine];
		strip(keyp, keybuf, keybufsize);
		currLine++;
		currLine++;
		break;
	case EZLM_FILE_TYPE_INI:
		if (!(keyp = strchr(filelines[currLine], (int)'='))) {
			printf("*** Invalid property file format - treating as EOF ***\n");
			return 0;
		}
		strip(keyp+1, keybuf, keybufsize);
		currLine++;
		break;
	case EZLM_FILE_TYPE_KEY:
		strip(filelines[currLine], keybuf, keybufsize);
		currLine++;
		break;
	case EZLM_FILE_TYPE_CSV:
	case EZLM_FILE_TYPE_TAB:
		keyp = strrchr(filelines[currLine], (fileType==EZLM_FILE_TYPE_CSV?(int)',':(int)'\t'));
		if (!keyp) {
			printf("*** Invalid %s file format - treating as EOF ***\n",
					(fileType==EZLM_FILE_TYPE_CSV?"csv":"tab"));
			return 0;
		}
		strip(keyp+1, keybuf, keybufsize);
		currLine++;
		break;
	}
	return currLine;
}


int CALLING_CONVENTION getFileType(const char *fname) {
	char *suffix = strrchr(fname, (int)'.');
	if (suffix > fname && suffix > &fname[strlen(fname)-4])
		return EZLM_FILE_TYPE_UNKNOWN;
	suffix++;
	if (strcmp(suffix, "txt")==0)
		return EZLM_FILE_TYPE_TXT;
	if (strcmp(suffix, "xml")==0)
		return EZLM_FILE_TYPE_XML;
	if (strcmp(suffix, "csv")==0)
		return EZLM_FILE_TYPE_CSV;
	if (strcmp(suffix, "tab")==0)
		return EZLM_FILE_TYPE_TAB;
	if (strcmp(suffix, "ini")==0)
		return EZLM_FILE_TYPE_INI;
	if (strcmp(suffix, "key")==0)
		return EZLM_FILE_TYPE_KEY;
	printf("*** Unknown file type: '%s'.  Exiting ***\n", suffix);
	exit(1);
}

/* 
 * Strip the key of all leading, trailing and embedded white spaces.
 * A good idea for an actual application, too.
 */
int CALLING_CONVENTION strip(const char *instr, char retbuf[], size_t retbufsize)
{
	char *outptr = &retbuf[0];
	int len = 0;
	char ch = 0;
	if (retbufsize <= strlen(instr))
		return -1;
	while (ch = *instr++) {
		if (!isspace((int)ch))
			*outptr++ = ch;
	}
	*outptr = '\0';
	return 0;
}


void makeDateStr(time_t ltime, char dateStrBuf[], size_t dateStrBufsize)
{
	char endch;
	struct tm when;
	if (ltime == 0)
		time(&ltime);
	when = *gmtime(&ltime);
	strncpy(dateStrBuf, asctime(&when), dateStrBufsize-1);
	endch = dateStrBuf[strlen(dateStrBuf)-1];
	if (iscntrl((int)endch))
		dateStrBuf[strlen(dateStrBuf)-1]='\0';
}

